from __future__ import unicode_literals

from typing import Any

from nornir.core.task import Result, Task


def netmiko_commit(task: Task, **kwargs: Any) -> Result:
    """
    Execute Netmiko commit method

    Arguments:
        kwargs: Additional arguments to pass to method.

    Returns:
        :obj: `nornir.core.task.Result`:
          * result (``str``): String showing the CLI output from the commit operation
    """
    conn = task.host.get_connection("netmiko", task.nornir.config)
    result = conn.commit(**kwargs)
    return Result(host=task.host, result=result, changed=True)



import logging
import threading
from typing import Union, cast

from colorama import Fore, Style, init

from nornir.core.inventory import Host
from nornir.core.task import AggregatedResult, MultiResult, Result, Task

init(autoreset=True, strip=False)


def _get_color(result: Union[MultiResult, Result]) -> str:
    if result.failed:
        color = Fore.RED
    elif result.changed:
        color = Fore.YELLOW
    else:
        color = Fore.GREEN
    return cast(str, color)


class PrintResult:
    """
    Prints information about the task execution on screen.

    Arguments:
        severity_level: Print only results with this severity level or higher
    """

    def __init__(self, severity_level: int = logging.INFO) -> None:
        self.severity_level = severity_level
        self.lock = threading.Lock()

    def task_started(self, task: Task) -> None:
        if task.severity_level < self.severity_level:
            return

        msg = f"**** {task.name} - Starting "
        print(f"{Style.BRIGHT}{Fore.CYAN}{msg}{'*' * (80 - len(msg))}")

    def task_completed(self, task: Task, result: AggregatedResult) -> None:
        if task.severity_level < self.severity_level:
            return

        msg = f"**** {task.name} - Completed "
        print(f"{Style.BRIGHT}{Fore.CYAN}{msg}{'*' * (80 - len(msg))}")

    def task_instance_started(self, task: Task, host: Host) -> None:
        pass

    def task_instance_completed(
        self, task: Task, host: Host, results: MultiResult
    ) -> None:
        self.lock.acquire()
        if task.severity_level < self.severity_level:
            self.lock.release()
            return

        # print task
        msg = f"vvvv {task.name} - {host.name} ** changed: {results.changed} "
        level_name = logging.getLevelName(results.severity_level)
        print(
            f"{Style.BRIGHT}{_get_color(results)}{msg}{'v' * (80 - len(msg))} {level_name}"
        )
        print(results.result)

        # print task footer
        print(f"{Style.BRIGHT}{_get_color(results)}{'^' * 80}")

        self.lock.release()

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        self.task_instance_started(task, host)

    def subtask_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        self.task_instance_completed(task, host, result)



import logging
import logging.config
from multiprocessing.dummy import Pool
from typing import List, Optional

from nornir.core.configuration import Config
from nornir.core.inventory import Inventory
from nornir.core.processor import Processor, Processors
from nornir.core.state import GlobalState
from nornir.core.task import AggregatedResult, Task

logger = logging.getLogger(__name__)


class Nornir(object):
    """
    This is the main object to work with. It contains the inventory and it serves
    as task dispatcher.

    Arguments:
        inventory (:obj:`nornir.core.inventory.Inventory`): Inventory to work with
        data(GlobalState): shared data amongst different iterations of nornir
        dry_run(``bool``): Whether if we are testing the changes or not
        config (:obj:`nornir.core.configuration.Config`): Configuration object

    Attributes:
        inventory (:obj:`nornir.core.inventory.Inventory`): Inventory to work with
        data(:obj:`nornir.core.GlobalState`): shared data amongst different iterations of nornir
        dry_run(``bool``): Whether if we are testing the changes or not
        config (:obj:`nornir.core.configuration.Config`): Configuration parameters
    """

    def __init__(
        self,
        inventory: Inventory,
        config: Config = None,
        data: GlobalState = None,
        processors: Optional[Processors] = None,
    ) -> None:
        self.data = data if data is not None else GlobalState()
        self.inventory = inventory
        self.config = config or Config()
        self.processors = processors or Processors()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_connections(on_good=True, on_failed=True)

    def with_processors(self, processors: List[Processor]) -> "Nornir":
        """
        Given a list of Processor objects return a copy of the nornir object with the processors
        assigned to the copy. The orinal object is left unmodified.
        """
        return Nornir(**{**self.__dict__, **{"processors": Processors(processors)}})

    def filter(self, *args, **kwargs):
        """
        See :py:meth:`nornir.core.inventory.Inventory.filter`

        Returns:
            :obj:`Nornir`: A new object with same configuration as ``self`` but filtered inventory.
        """
        b = Nornir(**self.__dict__)
        b.inventory = self.inventory.filter(*args, **kwargs)
        return b

    def _run_serial(self, task: Task, hosts, **kwargs):
        result = AggregatedResult(kwargs.get("name") or task.name)
        for host in hosts:
            result[host.name] = task.copy().start(host, self)
        return result

    def _run_parallel(self, task: Task, hosts, num_workers, **kwargs):
        result = AggregatedResult(kwargs.get("name") or task.name)

        pool = Pool(processes=num_workers)
        result_pool = [
            pool.apply_async(task.copy().start, args=(h, self)) for h in hosts
        ]
        pool.close()
        pool.join()

        for rp in result_pool:
            r = rp.get()
            result[r.host.name] = r
        return result

    def run(
        self,
        task,
        num_workers=None,
        raise_on_error=None,
        on_good=True,
        on_failed=False,
        **kwargs,
    ):
        """
        Run task over all the hosts in the inventory.

        Arguments:
            task (``callable``): function or callable that will be run against each device in
              the inventory
            num_workers(``int``): Override for how many hosts to run in parallel for this task
            raise_on_error (``bool``): Override raise_on_error behavior
            on_good(``bool``): Whether to run or not this task on hosts marked as good
            on_failed(``bool``): Whether to run or not this task on hosts marked as failed
            **kwargs: additional argument to pass to ``task`` when calling it

        Raises:
            :obj:`nornir.core.exceptions.NornirExecutionError`: if at least a task fails
              and self.config.core.raise_on_error is set to ``True``

        Returns:
            :obj:`nornir.core.task.AggregatedResult`: results of each execution
        """
        task = Task(task, **kwargs)
        self.processors.task_started(task)

        num_workers = num_workers or self.config.core.num_workers

        run_on = []
        if on_good:
            for name, host in self.inventory.hosts.items():
                if name not in self.data.failed_hosts:
                    run_on.append(host)
        if on_failed:
            for name, host in self.inventory.hosts.items():
                if name in self.data.failed_hosts:
                    run_on.append(host)

        num_hosts = len(self.inventory.hosts)
        task_name = kwargs.get("name") or task.name
        if num_hosts:
            logger.info(
                f"Running task %r with args %s on %d hosts",
                task_name,
                kwargs,
                num_hosts,
            )
        else:
            logger.warning("Task %r has not been run – 0 hosts selected", task_name)

        if num_workers == 1:
            result = self._run_serial(task, run_on, **kwargs)
        else:
            result = self._run_parallel(task, run_on, num_workers, **kwargs)

        raise_on_error = (
            raise_on_error
            if raise_on_error is not None
            else self.config.core.raise_on_error
        )  # noqa
        if raise_on_error:
            result.raise_on_error()
        else:
            self.data.failed_hosts.update(result.failed_hosts.keys())

        self.processors.task_completed(task, result)

        return result

    def dict(self):
        """ Return a dictionary representing the object. """
        return {"data": self.data.dict(), "inventory": self.inventory.dict()}

    def close_connections(self, on_good=True, on_failed=False):
        def close_connections_task(task):
            task.host.close_connections()

        self.run(task=close_connections_task, on_good=on_good, on_failed=on_failed)

    @classmethod
    def get_validators(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, cls):
            raise ValueError(f"Nornir: Nornir expected not {type(v)}")
        return v

    @property
    def state(self):
        return GlobalState



from typing import Any, Dict, Optional

from ncclient import manager

from nornir.core.configuration import Config
from nornir.core.connections import ConnectionPlugin


class Netconf(ConnectionPlugin):
    """
    This plugin connects to the device via NETCONF using ncclient library.

    Inventory:
        extras: See
        `here <https://ncclient.readthedocs.io/en/latest/transport.html#ncclient.transport.SSHSession.connect>`_

    Example on how to configure a device to use netconfig without using an ssh agent and without verifying the keys::

        ---
        nc_device:
            hostname: 192.168.16.20
            username: admin
            password: admin
            port: 2022
            connection_options:
                netconf:
                    extras:
                        allow_agent: False
                        hostkey_verify: False

    Then it can be used like::

        >>> from nornir import InitNornir
        >>> from nornir.core.task import Result, Task
        >>>
        >>> nr = InitNornir(
        >>>    inventory={
        >>>        "options": {
        >>>            "hosts": {
        >>>                "rtr00": {
        >>>                    "hostname": "localhost",
        >>>                   "username": "admin",
        >>>                    "password": "admin",
        >>>                    "port": 65030,
        >>>                    "platform": "whatever",
        >>>                    "connection_options": {
        >>>                        "netconf": {"extras": {"hostkey_verify": False}}
        >>>                    },
        >>>                }
        >>>           }
        >>>        }
        >>>    }
        >>>)
        >>>
        >>>
        >>> def netconf_code(task: Task) -> Result:
        >>>    manager = task.host.get_connection("netconf", task.nornir.config)
        >>>
        >>>    # get running config and system state
        >>>    print(manager.get())
        >>>
        >>>    # get only hostname
        >>>    print(manager.get(filter=("xpath", "/sys:system/sys:hostname")))
        >>>
        >>>    # get candidate config
        >>>    print(manager.get_config("candidate"))
        >>>
        >>>    # lock
        >>>    print(manager.lock("candidate"))
        >>>
        >>>    # edit configuration
        >>>    res = manager.edit_config(
        >>>        "candidate",
        >>>        "<sys:system><sys:hostname>asd</sys:hostname></sys:system>",
        >>>        default_operation="merge",
        >>>    )
        >>>    print(res)
        >>>
        >>>    print(manager.commit())
        >>>
        >>>    # unlock
        >>>    print(manager.unlock("candidate"))
        >>>
        >>>    return Result(result="ok", host=task.host)
        >>>
        >>>
        >>> nr.run(task=netconf_code)
    """  # noqa

    def open(
        self,
        hostname: Optional[str],
        username: Optional[str],
        password: Optional[str],
        port: Optional[int],
        platform: Optional[str],
        extras: Optional[Dict[str, Any]] = None,
        configuration: Optional[Config] = None,
    ) -> None:
        extras = extras or {}

        parameters: Dict[str, Any] = {
            "host": hostname,
            "username": username,
            "password": password,
            "port": port or 830,
        }

        if "ssh_config" not in extras:
            try:
                parameters["ssh_config"] = configuration.ssh.config_file  # type: ignore
            except AttributeError:
                pass

        parameters.update(extras)
        parameters["ssh_config"] = None

        connection = manager.connect_ssh(**parameters)
        self.connection = connection

    def close(self) -> None:
        self.connection.close_session()
